Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div choice fintype.
Require Import finfun finset ssralg perm zmodp seqmatrix1.
Require Import Recdef.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(*In this file is the definition of an admissible ordered discrete vector field.
  So, if vf is a ordered vector field from M and the relations satisfy the admissibility property then Vecfield M vf ors will be true *)



Section definitions.

Local Open Scope ring_scope.

(* We define the ring Z2 *)
Definition Z2 := Fp_fieldType 2.

Definition matZ2 := seq (seq Z2).
Definition seqZ2 := seq Z2.

(*Definition of vectorfield is a list of pairs of natural numbers*)
Definition vectorfield := seq (prod nat nat).

Definition orders:= seq (seq nat).
Definition order:= seq nat.

Definition is_matrix m n (M:matZ2) := 
   M = [::] \/ [/\ m = size M & forall i, i < m -> size (rowseqmx M i) = n].


Section sequences.

(*return a list with the first components of the sequence of pairs*)
Definition getfirstElemseq (s: seq (nat * nat)):= map (@fst _ _) s.

(*return a list with the second components of the sequence of pairs*)
Definition getsndElemseq (s: seq (nat*nat)):= map (@snd _ _) s.

(*check every element of the sequence is higher than 0 and lower than m*)
Definition longmn m:= all [pred i | 0<= i < m].


End sequences.


Section functions_lemmas.

(*this function is necessary to prove the property of admissibility*)
(*check every list of a list of list of naturals hasn't repeated*)
Definition norep (rest: orders):= all uniq rest.

(*return the value of the entry (i,j) of the matrix M*)
Definition compij (i j:nat)(M: matZ2):= nth 0 (nth nil M i) j.




(*to order a dvf according to f (higher to lower) *)
Section insert_sortf.

Variable f: nat -> orders -> nat.

Fixpoint ordered (s:seq (nat*nat)) ords:=
match s with
|nil => true
|a::b => match b with
        |nil => true
        |b1::b2 => ((f (fst a) ords) >= (f (fst b1) ords)) && ordered b ords
        end
end.



Fixpoint insert n (s:seq (nat * nat)) (ord: orders) :=
match s with
nil => [::n]
|a::b => if ((f (fst a) ord) <= (f (fst n) ord)) 
         then n::(a::b)
         else a::(insert n b ord)
end.

Fixpoint insert_sort (s : seq (nat * nat)) (ord: orders):=
match s with
nil => nil
|a::b => insert a (insert_sort b ord) ord 
end.

Functional Scheme insert_ind := Induction for insert Sort Prop.


Lemma ordered_insert s n ords: ordered s ords -> ordered (insert n s ords) ords.
Proof.
elim H: s=>[|s1 s2]; first by rewrite /=.
move=> HI.
rewrite /ordered -/ordered.
case H1: s2=>[|s21 s22]. rewrite /insert.
  case H2:(f s1.1 ords <= f n.1 ords); first by rewrite /ordered H2.
  rewrite /ordered.
  move:H2; move/negP; move/negP; rewrite -ltnNge.
  move=> H2 _.
  by rewrite leq_eqVlt Bool.andb_true_r; apply/orP; right.
rewrite -H1.
move/andP=>[H2 H3].
move: (HI H3)=> H4.
rewrite /insert -/insert.
case H5:(f s1.1 ords <= f n.1 ords).
  by rewrite /ordered  H5 -/ordered H1 H2 -H1 H3.
rewrite /ordered -/ordered.
case H6:(insert n s2 ords)=>[|t s3]; first by done.
rewrite -H6 H4 /=.
move: H6; rewrite /insert  H1 -/insert.
case H6:(f s21.1 ords <= f n.1 ords).
  move/eqP; rewrite eqseq_cons; move/andP=>[H7 H8].
  move: (eqP H7)=> <-.
  move:H5; move/negP; move/negP; rewrite -ltnNge.
  move=> H21.
  by rewrite leq_eqVlt Bool.andb_true_r; apply/orP; right.
move/eqP; rewrite eqseq_cons; move/andP=>[H7 H8].
by move: (eqP H7)=> <-; rewrite H2.
Qed.



Lemma ordered_insert_sort s ords: ordered (insert_sort s ords) ords.
Proof.
elim H: s=>[|s1 s2]; first by rewrite /=.
move=> HI.
rewrite /insert_sort -/insert_sort.
apply: (ordered_insert s1 HI).
Qed.


End insert_sortf.



Close Local Scope ring_scope.

Fixpoint lengthMax1 (ords: orders):= 
match ords with
|nil => 0 
|a::b => maxn (maxn 0 (size a)) (lengthMax1 b)
end.

(* return the sequences which start with k *)
Definition getfirstE (k:nat)(ord: orders):= filter [pred s | head 0%N s == k] ord.

Definition glMax (n:nat)(ord:orders):=  lengthMax1 (getfirstE n ord).

Definition prop_cat2 (ords: seq (seq nat)):= 
  (forall a b s p, (a::s) \in ords -> (last 0%N s = b) -> (b::p) \in ords -> ((a::s)++p) \in ords).


Definition prop_cat (ords: seq (seq nat)):= 
  all [pred x| (all [pred y|  (x++ (behead y)) \in ords] (filter [pred y | last 0%N (behead x) == head 0%N y] ords))] ords.


Lemma prop_cat_prop_cat2 ords: (prop_cat ords) -> (prop_cat2 ords).
Proof.
move/allP. rewrite /=.
rewrite /prop_cat =>[H1 a b s p H4 H2 H3].
move: (H1 (a::s) H4).
move/allP; rewrite /=.
move=> H5.
have H6: ((b::p) \in filter [pred y | last 0%N s == head 0%N y] ords).
  rewrite mem_filter.
  by rewrite H3 /= Bool.andb_true_r H2.
move: (H5 (b::p) H6).
by rewrite /=.
Qed.


End functions_lemmas.





Section definition_dvfadm.

Variable ss: matZ2.
Variable m1 n1:nat.
Hypotheses ismatrix: is_matrix m1 n1 ss.


Definition Vecfieldadm (ss: matZ2)(vf: vectorfield)(ord:orders) :=
  (longmn (size ss) (getfirstElemseq vf)) /\ 
  (longmn (size (nth [::] ss 0)) (getsndElemseq vf)) /\
  (uniq (getfirstElemseq vf)) /\ (uniq (getsndElemseq vf)) /\ 
  (forall i j:nat, (i , j) \in vf -> compij i j ss = 1%R) /\
  (norep ord) /\
  (ordered glMax vf ord) /\
  (forall i j m:nat, (i,j)\in vf -> i!=m -> compij m j ss != 0%R -> (i::m::nil)\in ord) /\
  (prop_cat2 ord).

Definition Vecfieldadm2 (ss: matZ2)(vf: vectorfield)(ord:orders) :=
  (longmn (size ss) (getfirstElemseq vf)) /\  
  (longmn (size (nth [::] ss 0)) (getsndElemseq vf)) /\
  (uniq (getfirstElemseq vf)) /\ (uniq (getsndElemseq vf)) /\ 
  (forall i j:nat, (i , j) \in vf -> compij i j ss = 1%R) /\
  (norep ord) /\
  (ordered glMax vf ord) /\
  (forall i j m:nat, (i,j)\in vf -> i!=m -> compij m j ss != 0%R -> m < size ss -> (i::m::nil)\in ord) /\
  (prop_cat2 ord).



End definition_dvfadm.

End definitions.
