Add LoadPath "CoqEAL".
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.
Require Import reduction_definition_two triangular triangular_upper.
Require Import inverse_nilpotent  ZArith Zinfra. 
Require Import BPL redCC2_general definitions_auxlemmas.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section Vector_Field_Reduction_Theorem.

Local Open Scope ring_scope.

(*We prove the vector-field reduction theorem. 
  
  We have a chain complex D (D1 and D2) with D1*D2 = 0 and the upper-left matrix of
  D1 is an upper triangular matrix.
*)

Section initial_chain_complex.

Variable K : fieldType.
Variables (m1 m2 n2 l : nat).
Variables (D1 : 'M[K]_(m1+m2,m1+n2)) (D2 : 'M[K]_(l,m1+m2)).
Hypothesis nilpotent_D1D2 : D2 *m D1 = 0.

Definition daa := (ulsubmx D1).
Definition dba := (ursubmx D1).
Definition dab := (dlsubmx D1).
Definition dbb := (drsubmx D1).
Definition d2a := (lsubmx D2).
Definition d2b := (rsubmx D2).


Lemma decomposition_D1 : D1 =
  block_mx daa dba dab dbb.
Proof.
by rewrite -(submxK D1). 
Qed.

Lemma decomposition_D2 : D2 = row_mx d2a d2b.
Proof.
by rewrite -(hsubmxK D2). 
Qed.


Hypothesis daa_upper_triangular : upper_triangular daa.



(*We define the initial reduction. The chain complex is (delta1, 0) and the reduced 
  (0 ,0) with the corresponding fi, gi and hi.*)

Section reduction_for_BPL.

Definition delta1 := block_mx 1%:M 0 0 0 : 'M[K]_(m1+m2,m1+n2).

Definition trivial1 := 0 : 'M[K]_(m2,n2).

Definition f0 := (col_mx 0 1%:M) : 'M[K]_(m1+n2,n2).
Definition f1 := (col_mx 0 1%:M) : 'M[K]_(m1+m2,m2).

Definition g0 := (row_mx 0 1%:M) : 'M[K]_(n2,m1+n2).
Definition g1 := (row_mx 0 1%:M) : 'M[K]_(m2,m1+m2).

Definition h0 := delta1^T.

Lemma f0D1_d1f1 : delta1 *m f0 = f1 *m trivial1.
Proof.
by rewrite mul_block_col !mul0mx !mulmx0 !GRing.addr0 col_mx0.
Qed.

Lemma g0d1_D1g1 : trivial1 *m g0 = g1 *m delta1.
Proof.
by rewrite mul_row_block !mulmx0 !mul0mx !GRing.addr0 row_mx0.
Qed.

Lemma f0g0 : g0 *m f0 = 1%:M.
Proof. by rewrite mul_row_col mulmx1 mulmx0 GRing.add0r. Qed.

Lemma f1g1 : g1 *m f1 = 1%:M.
Proof. by rewrite mul_row_col mulmx1 mulmx0 GRing.add0r. Qed.

Lemma g0f0_D1h0 : f0 *m g0 + h0 *m delta1 = 1%:M.
Proof.
by rewrite mul_col_row /h0 /delta1 !mulmx1 mul0mx mulmx0 tr_block_mx !trmx0 
  !trmx1 mulmx_block add_block_mx !GRing.add0r !mulmx1 !mul1mx !mul0mx
  !GRing.addr0 -scalar_mx_block.  
Qed.

Lemma g1f1_D2h1_D1h0 : f1 *m g1 + (H1 K (m1 + m2) l) *m 
                                  (redCC2_general.D2 K (m1+m2) l) +
                       delta1 *m h0 = 1%:M.
Proof.
by rewrite mulmx0 !GRing.addr0 mul_col_row !mul0mx mulmx0 mul1mx
       /delta1 /h0 tr_block_mx !trmx0 
       !trmx1 mulmx_block add_block_mx !GRing.add0r !mulmx1 !mul1mx !mul0mx
       !GRing.addr0 -scalar_mx_block.  
Qed.

Lemma f1h0 : h0 *m f1 = 0.
Proof.
by rewrite /h0 tr_block_mx !trmx0 !trmx1 mul_block_col !mul0mx !mulmx0 
        !GRing.addr0 col_mx0.
Qed.

Lemma h0g0 : g0 *m h0 = 0.
Proof.
by rewrite /h0 tr_block_mx !trmx0 !trmx1 mul_row_block !mul0mx !mulmx0
   !GRing.addr0 row_mx0.
Qed.


Definition D3 : 'M[K]_(0,l) := 0.
Definition D0 : 'M[K]_(m1+n2,0):=0.
Definition Dn : 'M[K]_(0,0):=0.


Definition D3' : 'M[K]_(0,l):= 0.
Definition D0' : 'M[K]_(n2,0) := 0.
Definition Dn' : 'M[K]_(0,0) := 0.

Definition Fn : 'M[K]_(0,0):=0.

Definition Gn : 'M[K]_(0,0):=0.

Definition H2 : 'M[K]_(l,0):=0.
Definition H00 : 'M[K]_(0,m1+n2):=0.
Definition Hn : 'M[K]_(0,0):=0.


(*We have defined a general reduction given by the definition FGReduction*)
Definition reductionFG:= (red_FGReduction D3 D0 Dn D3' D0' Dn' f0D1_d1f1 g0d1_D1g1 f0g0 f1g1 g0f0_D1h0 g1f1_D2h1_D1h0 f1h0 h0g0 Fn Gn H2 H00 Hn).  



(*To apply BPL, we define two perturbations (P1 = D1 - delta1 and P2 = D2)
  In this way, the initial perturbed chain complex is composed of (D1, D2).

  Then, applying BPL we can define a reduction reduction_vector_field between 
   the chain complex perturbed which is exactly the same that the initial.
*)

Definition P1 := D1 - delta1.
Definition P2 := D2.

Definition P3 : 'M[K]_(0,l):=0.
Definition P0 : 'M[K]_(m1+n2,0):=0.
Definition Pn : 'M[K]_(0,0):=0.


Definition m_m1: Z -> nat:= fun i:Z => match i with
|0%Z => (m1+n2)%nat
|1%Z => (m1+m2)%nat
|2%Z => l%nat
|_ => 0%nat
end.


Definition P_m:= fun i:Z => match i as
 z return ((fun z0 : Z => 'M_(m_m1 (z0 + 1), m_m1 z0)) z) with
|(-1)%Z => P0
|0%Z => P1
|1%Z => P2
|2%Z => P3
|3%Z => Pn
|_ => Pn
end.


(*first property of BPL: nilpotency differential*)

Lemma m_m_0 (R : ringType) (t1 t2 : nat) (M : 'M[R]_(t1,t2)) : M - M = 0.
Proof.
by move : (eq_refl M); rewrite -GRing.subr_eq0; move/eqP => ->.
Qed.


Lemma nilpotent_d1d2_pert :
  (redCC2_general.D2 K (m1 + m2) l + P2) *m (delta1 + P1) = 0.
Proof.
by rewrite /redCC2_general.D2 /P2 /P1 GRing.add0r GRing.addrA
        [delta1 + _]GRing.addrC -GRing.addrA m_m_0 GRing.addr0
        nilpotent_D1D2.
Qed.



Lemma nilpotencyD0P0n: (D0 + P0) *m (Dn + Pn) = 0.
Proof.
by rewrite !thinmx0.
Qed.


Lemma nilpotencyD1P10:(delta1 + P1) *m (D0 + P0) = 0.
Proof.
by rewrite !thinmx0.
Qed.


Lemma nilpotencyDnPn: (Dn + Pn) *m (Dn + Pn) = 0.
Proof.
by rewrite !thinmx0.
Qed.

Lemma nilpotencyDnPn3: (Dn + Pn) *m (D3 + P3) = 0.
Proof.
by rewrite !thinmx0 mul0mx.
Qed.


Lemma nilpotencyD3P3D2P2:(D3 + P3) *m (redCC2_general.D2 K (m1 + m2) l + P2) = 0.
Proof.
by rewrite !flatmx0.
Qed.


Lemma nilpotency_dp_m: forall i : Z,
        (diff (C reductionFG) (i + 1) + P_m (i + 1)) *m 
        (diff (C reductionFG) i + P_m i) = 0.
Proof.
move=> i.
case: i.
  rewrite /=.
  exact: nilpotent_d1d2_pert.
move=> p.
case:p.
  move=> p. 
  case:p; rewrite /=; try by move=> k; exact nilpotencyDnPn.
  exact: nilpotencyDnPn.
  move=> p.
  case:p; rewrite /=; try by move=> k; exact nilpotencyDnPn.
  exact: nilpotencyDnPn3.
rewrite /=; exact: nilpotencyD3P3D2P2.
move=>p.
case:p.
move=> p; rewrite /=; try by move=> k; exact nilpotencyDnPn. 
exact nilpotencyDnPn.
move=> p.
case:p; rewrite /=; try by move=> k; exact nilpotencyDnPn.
exact nilpotencyD0P0n. 
rewrite /=; exact nilpotencyD1P10.
Qed.



(*second property: nilpotency (P*H)*)

Lemma ul_pot_block_a_0_b_0 (p1 p2 k: nat) (a : 'M[K]_p1)
   (b : 'M[K]_(p2,p1)) : 
  ulsubmx (pot_matrix (block_mx a 0 b 0) k) = pot_matrix a k.
Proof.
elim : k =>  [/=| k H].
  by rewrite scalar_mx_block block_mxKul.
have -> : (pot_matrix (block_mx a 0 b 0) k.+1) =
 (block_mx a 0 b 0) *m (pot_matrix (block_mx a 0 b 0) k) by done.
by rewrite -(submxK (pot_matrix (block_mx a 0 b 0) k)) mulmx_block
        block_mxKul mul0mx H GRing.addr0.
Qed.

Lemma ur_pot_block_a_0_b_0 (p1 p2 k: nat) (a : 'M[K]_p1)
   (b : 'M[K]_(p2,p1)) : 
  ursubmx (pot_matrix (block_mx a 0 b 0) k) = 0.
Proof.
elim : k =>  [/=| k H].
  by rewrite scalar_mx_block block_mxKur.
have -> : (pot_matrix (block_mx a 0 b 0) k.+1) =
 (block_mx a 0 b 0) *m (pot_matrix (block_mx a 0 b 0) k) by done.
by rewrite -(submxK (pot_matrix (block_mx a 0 b 0) k)) mulmx_block
        block_mxKur mul0mx H GRing.addr0 mulmx0.
Qed.


Lemma dl_pot_block_a_0_b_0 (p1 p2 k: nat) (a : 'M[K]_p1)
   (b : 'M[K]_(p2,p1)) : 0 < k ->
  dlsubmx (pot_matrix (block_mx a 0 b 0) k) = b *m pot_matrix a k.-1.
Proof.
elim : k =>  [/=| k H1 H22].
  by done.
have -> : (pot_matrix (block_mx a 0 b 0) k.+1) =
 (block_mx a 0 b 0) *m (pot_matrix (block_mx a 0 b 0) k) by done.
have : 0 < k \/ k = 0%N.
  move : H1 H22.
  case : k; first by right.
  by left.
move => [Hnn|->]; last first.
  by rewrite mulmx1 -pred_Sn mulmx1 block_mxKdl.
rewrite -pred_Sn.
by rewrite -(submxK (pot_matrix (block_mx a 0 b 0) k)) mulmx_block
        block_mxKdl mul0mx ul_pot_block_a_0_b_0 GRing.addr0. 
Qed.


Lemma dr_pot_block_a_0_b_0 (p1 p2 k: nat) (a : 'M[K]_p1)
   (b : 'M[K]_(p2,p1)) : 0 < k ->
  drsubmx (pot_matrix (block_mx a 0 b 0) k) = 0.
Proof.
elim : k =>  [/=| n H1 H22].
  by done.
have -> : (pot_matrix (block_mx a 0 b 0) n.+1) =
 (block_mx a 0 b 0) *m (pot_matrix (block_mx a 0 b 0) n) by done.
have : 0 < n \/ n = 0%N.
  move : H1 H22.
  case : n; first by right.
  by left.
move => [Hnn|->].
  by rewrite  -(submxK (pot_matrix (block_mx a 0 b 0) n)) mulmx_block
        block_mxKdr (H1 Hnn) ur_pot_block_a_0_b_0 !mulmx0 GRing.addr0.
by rewrite mulmx1 block_mxKdr.
Qed.


Lemma nilpotent_hp1 :
  pot_matrix (P1 *m h0) (m1.+1).+1 = 0.
Proof.
rewrite /P1 opp_block_mx decomposition_D1 add_block_mx !GRing.subr0/h0.
rewrite tr_block_mx !trmx0 !trmx1 mulmx_block !mulmx0 !GRing.addr0 
       !mulmx1. rewrite -(submxK (pot_matrix (block_mx (daa - 1%:M) 0 dab 0) 
                                    (m1.+1).+1)).
have m10 : 0 < (m1.+1).+1 by done.
rewrite (@dr_pot_block_a_0_b_0 _ _ _ _ _ m10) 
        (@dl_pot_block_a_0_b_0 _ _ _ _ _ m10)
        ur_pot_block_a_0_b_0 ul_pot_block_a_0_b_0 -pred_Sn.
have -> : (pot_matrix (daa - 1%:M) m1.+2) =
          (daa - 1%:M) *m (pot_matrix (daa - 1%:M) m1.+1) by done.
have -> : (pot_matrix (daa - 1%:M) m1.+1) = 0; last first.
  by rewrite !mulmx0 block_mx0.
have H : upper_triangular_up_to_k 0 (daa - 1%:M)
  by apply : (upper_triangular_minus_1 daa_upper_triangular).
by exact (upper_triangular_pot_matrix_n H).
Qed.



Lemma nilpotency_hpn : (pot_matrix (Pn *m Hn)  ((m1.+1).+1) = 0).
Proof.
by rewrite !thinmx0.
Qed.


Lemma nilpotency_hp3: (pot_matrix (P_m 3 *m Ho (H reductionFG) 3) m1.+2) = 0.
Proof.
by rewrite !thinmx0.
Qed.


Lemma nilpotency_hp32: (pot_matrix (P3 *m H2) m1.+2) = 0.
Proof.
by rewrite !thinmx0.
Qed.


Lemma nilpotency_hp11: (pot_matrix (P_m 1 *m Ho (H reductionFG) 1) m1.+2) = 0.
Proof.
by rewrite /= /P2 /H1 mulmx0 mul0mx.
Qed.


Lemma nilpotency_hp00: (pot_matrix (P0 *m H00) m1.+2) = 0.
Proof.
by rewrite !thinmx0 /= !mul0mx.
Qed.


Lemma nilpotency_hp_m :
   forall i:Z, (pot_matrix (P_m i *m (Ho (H reductionFG) i)) (m1.+1).+1 = 0).
Proof.
move=> i.
case:i.
  rewrite /=.
  exact: nilpotent_hp1.
move=> p.
case:p.
  move=> p. 
  case:p. move=> k; exact nilpotency_hpn.
  move=> p; exact nilpotency_hpn.
  exact: nilpotency_hp3.
  move=> p.
  case:p; rewrite /=; try by exact nilpotency_hpn.
  move=> k; exact: nilpotency_hpn.
move=>p; exact: nilpotency_hpn.
exact: nilpotency_hp32.
exact: nilpotency_hp11.
move=> p.
case:p; rewrite /=; try by move=> k; exact nilpotency_hpn.
exact nilpotency_hp00. 
Qed.



(*applying general basic perturbation lemma*)
Definition reduction_BPL_general := (rho_def nilpotency_dp_m nilpotency_hp_m).


End reduction_for_BPL.

End initial_chain_complex.

End Vector_Field_Reduction_Theorem.
