(* This file is part of CoqEAL, the Coq Effective Algebra Library *)
Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq choice fintype.
Require Import div finfun bigop prime binomial ssralg finset fingroup finalg.
Require Import perm zmodp matrix.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section matrix_complements.

Variable R : ringType.


Lemma row_flat_mx (m n : nat) (M : 'M_(m,0)) (N : 'M[R]_(m,n)) :
  row_mx M N = N.
Proof.
apply/matrixP=> i j; rewrite mxE; case:(splitP j)=> [|k H]; first by case.
by congr fun_of_matrix; exact:val_inj.
Qed.


End matrix_complements.

Section seqmx.

Local Open Scope ring_scope.

Variable mT : ringType.

Section SeqmxDef.

Definition seqmatrix := seq (seq mT).
Definition seqrow := seq mT.

Variables m n : nat.

Definition mkseqmx (f:nat->nat->mT) : seqmatrix :=
  mkseq (fun i => mkseq (fun j => f i j) n) m.

Definition ord_enum_eq n : seq 'I_n := pmap (insub_eq _) (iota 0 n).

Lemma ord_enum_eqE p : ord_enum_eq p = enum 'I_p.
Proof. by rewrite enumT unlock; apply:eq_pmap ; exact:insub_eqE. Qed.


Definition mkseqmx_ord (f:'I_m->'I_n->mT) : seqmatrix :=
  let enum_n := ord_enum_eq n in
  map (fun i => map (f i) enum_n) (ord_enum_eq m).

Lemma eq_mkseqmx_ord f g :
  f =2 g -> mkseqmx_ord f = mkseqmx_ord g.
Proof. by move=> Hfg; apply:eq_map=> i; apply:eq_map=> j. Qed.

Definition rowseqmx (M : seqmatrix) i := nth [::] M i.

Definition mx_of_seqmx (M:seqmatrix) :=
  \matrix_(i < m, j < n) nth 0%R (rowseqmx M i) j.

Definition fun_of_seqmx (M:seqmatrix) := fun i j => nth 0%R (rowseqmx M i) j.

Coercion fun_of_seqmx : seqmatrix >-> Funclass.

Lemma fun_of_seqmxE (M:seqmatrix) i j : M i j = nth 0%R (nth [::] M i) j.
Proof. by []. Qed.


Lemma mkseqmxE f i j : i < m -> j < n -> mkseqmx f i j = f i j.
Proof. by move=> Hi Hj ; rewrite /fun_of_seqmx /rowseqmx !nth_mkseq. Qed.

Definition seqmx_of_mx (M:'M_(m,n)) : seqmatrix :=
  map (fun i => map (fun j => M i j) (enum 'I_n)) (enum 'I_m).

Lemma size_seqmx : forall (M:'M[mT]_(m,n)),
  size (seqmx_of_mx M) = m.
Proof. by move=> M ; rewrite size_map size_enum_ord. Qed.

Lemma seqmxE : forall (M:'M[mT]_(m,n)) (i:'I_m) (j:'I_n),
  (seqmx_of_mx M) i j = M i j.
Proof.
move=> M i j ; rewrite /fun_of_seqmx /rowseqmx.
rewrite (nth_map i); last by rewrite size_enum_ord ltn_ord.
rewrite (nth_map j); last by rewrite size_enum_ord ltn_ord.
by rewrite !nth_ord_enum.
Qed.

End SeqmxDef.

Section Degenerate.

Variables m n : nat.




Lemma seqmx0n (M:'M[mT]_(0,n)) : seqmx_of_mx M = [::].
Proof. by rewrite /seqmx_of_mx (size0nil (size_enum_ord 0)). Qed.

Lemma seqmxm0 (M:'M[mT]_(m,0)) : seqmx_of_mx M = map (fun _ => [::]) (enum 'I_m).
Proof. by rewrite /seqmx_of_mx (size0nil (size_enum_ord 0)). Qed.

End Degenerate.

Section FixedDim.

Variables m n : nat.

Lemma seqmx_eqP (M N : 'M[mT]_(m,n)) :
  reflect (M = N) (seqmx_of_mx M == seqmx_of_mx N).
Proof.
by apply/(iffP idP)=> [H|->//]; apply/matrixP=> i j; rewrite -!seqmxE (eqP H).
Qed.

Lemma size_row_seqmx : forall (M:'M[mT]_(m,n)) i,
  i < m -> size (rowseqmx (seqmx_of_mx M) i) = n.
Proof.
case:m=> [M i|p M i Hi] ; first by rewrite ltn0.
rewrite /rowseqmx (nth_map 0) ; first by rewrite size_map size_enum_ord.
by rewrite size_enum_ord.
Qed.

Lemma size_row_mkseqmx f i : 
  i < m -> size (rowseqmx (mkseqmx m n f) i) = n.
Proof. by move=> Hi ; rewrite /rowseqmx nth_mkseq // size_mkseq. Qed.

Lemma seqmxP (M : seqmatrix) (N : 'M_(m,n)) :
  [/\ m = size M, forall i, i < m -> size (rowseqmx M i) = n &
  forall (i:'I_m) (j:'I_n), M i j = N i j] <-> M = seqmx_of_mx N.
Proof.
split=> [[Hm Hn HMN]|->].
  apply:(@eq_from_nth _ [::])=> [|i Hi] ; first by rewrite size_seqmx.
  rewrite -Hm in Hi ; apply:(@eq_from_nth _ 0)=> [|j] ; rewrite (Hn _ Hi).
    by rewrite size_row_seqmx.
  by move=> Hj; rewrite -!fun_of_seqmxE (HMN (Ordinal Hi) (Ordinal Hj)) -seqmxE.
split=> [|i Hi|i j] ; first by rewrite size_seqmx.
  by rewrite size_row_seqmx.
by rewrite seqmxE.
Qed.

Lemma seqmx_of_funE (f : 'I_m -> 'I_n -> mT) :
  seqmx_of_mx (\matrix_(i < m, j < n) f i j) = mkseqmx_ord f.
Proof.
rewrite /mkseqmx_ord /seqmx_of_mx !ord_enum_eqE.
by apply:eq_map=> i; rewrite enumT unlock; apply:eq_map=> j; rewrite mxE.
Qed.

End FixedDim.

Section map2.

Variables T1 : Type.
Variables x1 : T1.
Variables T2 : Type.
Variable x2 : T2.
Variable T3 : Type.
Variable x3 : T3.
Variable (f : T1 -> T2 -> T3).

Fixpoint map2 (s1 : seq T1) (s2 : seq T2) :=
  if s1 is x1 :: s1' then
    if s2 is x2 :: s2' then f x1 x2 :: map2 s1' s2' else [::]
  else [::].

Lemma size_map2 (s1 : seq T1) (s2 : seq T2) :
  size (map2 s1 s2) = minn (size s1) (size s2).
Proof.
move:s2; elim:s1=>[s2|t1 s1 IHs] ; first by rewrite min0n.
by case=>[|t2 s2] //= ; rewrite IHs /minn /leq subSS ; case:ifP.
Qed.

Lemma nth_map2 (n : nat) (s1 : seq T1) (s2 : seq T2) :
n < minn (size s1) (size s2) ->
  nth x3 (map2 s1 s2) n = f (nth x1 s1 n) (nth x2 s2 n).
Proof.
move:n s2; elim:s1=> [n s2|t1 s1 IHs]; first by rewrite min0n ltn0.
case=>[|n]; first by case.
by case=> // t2 s2 H ; rewrite /= IHs // ; move:H ; rewrite !leq_minr.
Qed.

Fixpoint foldl2 (f : T3 -> T1 -> T2 -> T3) z (s : seq T1) (t : seq T2) {struct s} :=
  if s is x :: s' then
    if t is y :: t' then foldl2 f (f z x y) s' t' else z
  else z.

End map2.

Section SeqmxOp.

Variables m n p' : nat.
Local Notation p := p'.+1.

Definition map2seqmx (M N : seqmatrix) (f : mT -> mT -> mT) : seqmatrix :=
  map2 (map2 f) M N.

Lemma map2seqmxE (M N : 'M_(m,n)) (f : mT -> mT -> mT) :
  map2seqmx (seqmx_of_mx M) (seqmx_of_mx N) f = seqmx_of_mx (\matrix_(i < m, j < n) f (M i j) (N i j)).
Proof.
apply/seqmxP ; split=>[|i Hi|i j].
    by rewrite size_map2 !size_seqmx minnn.
  rewrite /rowseqmx (nth_map2 [::] [::]) ; last by rewrite !size_seqmx minnn.
  by rewrite size_map2 !size_row_seqmx // minnn.
rewrite fun_of_seqmxE (nth_map2 [::] [::]); last by rewrite !size_seqmx // minnn.
rewrite (nth_map2 0 0) ; last by rewrite !size_row_seqmx // minnn.
by rewrite -!fun_of_seqmxE !seqmxE mxE.
Qed.


Definition addseqmx (M N : seqmatrix) : seqmatrix :=
  map2seqmx M N (fun x y => x + y).

Lemma addseqmxE:
  {morph (@seqmx_of_mx m n) : M N / M + N >-> addseqmx M N}.
Proof. by move=> M N ; rewrite /addseqmx map2seqmxE. Qed.

Definition subseqmx (M N : seqmatrix) :=
  map2seqmx M N (fun x y => x - y).

Lemma subseqmxE:
  {morph (@seqmx_of_mx m n) : M N / M - N >-> subseqmx M N}.
Proof.
move=> M N ; rewrite /subseqmx map2seqmxE.
by congr seqmx_of_mx; apply/matrixP=> i j ; rewrite !mxE.
Qed.


Definition map_seqmx (f : mT -> mT) (M : seqmatrix) : seqmatrix :=
  map (map f) M.

Lemma map_seqmxE (M : 'M_(m,n)) (f :  mT -> mT) :
  map_seqmx f (seqmx_of_mx M) = seqmx_of_mx (\matrix_(i < m, j < n) f (M i j)).
Proof.
apply/seqmxP ; split=>[|i Hi|i j].
    by rewrite size_map !size_seqmx.
  rewrite /rowseqmx (nth_map [::]); last by rewrite !size_seqmx.
  by rewrite size_map !size_row_seqmx //.
rewrite fun_of_seqmxE (nth_map [::]); last by rewrite !size_seqmx.
rewrite (nth_map 0) ; last by rewrite !size_row_seqmx.
by rewrite -!fun_of_seqmxE !seqmxE mxE.
Qed.

Definition oppseqmx (M : seqmatrix) : seqmatrix :=
  map_seqmx (fun x => GRing.opp x) M.

Lemma oppseqmxE:
  {morph (@seqmx_of_mx m n) : M / - M >-> oppseqmx M}.
Proof. by move => M; rewrite /oppseqmx map_seqmxE. Qed.


Definition trseqmx (M : seqmatrix) : seqmatrix :=
  foldr (map2 cons) (nseq (size (rowseqmx M 0)) [::]) M.


Lemma size_trseqmx (M : 'M_(m.+1, n)) : size (trseqmx (seqmx_of_mx M)) = n.
Proof.
rewrite /trseqmx.
pose P s k := forall i, i < size s -> size (rowseqmx s i) = k.
have H: forall s1 s2, P s2 (size s1) -> size (foldr (map2 cons) s1 s2) = size s1.
  move=> s1; elim=> // t s2 IHs H.
  have Hs2: (P s2 (size s1)).
   by move=> i Hi; move:(H i.+1 Hi); move:(H 1%N (leq_ltn_trans (leq0n i) Hi))=> <-.
  by rewrite size_map2 (IHs Hs2) (H 0%N) // minnn.
rewrite H size_nseq size_row_seqmx //.
by move=> i; rewrite size_seqmx=> Hi; rewrite size_row_seqmx.
Qed.

Lemma size_row_trseqmx (M : 'M_(m.+1, n)) i :
  i < n -> size (rowseqmx (trseqmx (seqmx_of_mx M)) i) = m.+1.
Proof.
have H: forall k s1 s2, k < size (foldr (map2 cons) s1 s2) -> k < size s1 ->
size (rowseqmx (foldr (map2 cons) s1 s2) k) = (size s2 + size (rowseqmx s1 k))%N.
  move=> k s1; elim=> // t s2 IHs H1 H2.
  rewrite /rowseqmx (nth_map2 0 [::]).
    by rewrite /= IHs //; move:H1; rewrite size_map2 leq_minr; case/andP.
  by rewrite -(size_map2 cons).
move=> Hi; rewrite H.
  + by rewrite size_seqmx /rowseqmx nth_nseq; case:ifP.
  + by rewrite size_trseqmx.
  by rewrite size_nseq size_row_seqmx.
Qed.

Lemma trseqmxE (M : 'M_(m.+1, n)) :
  trseqmx (seqmx_of_mx M) = seqmx_of_mx (trmx M).
Proof.
apply/seqmxP; split => [|i Hi|i j].
  + by rewrite size_trseqmx.
  + by rewrite size_row_trseqmx.
rewrite /trseqmx /fun_of_seqmx.
have ->: forall s2 k l s1, k < size (foldr (map2 cons) s1 s2) -> (* -> l < size (rowseqmx s1 k) ->*)
nth 0 (nth [::] (foldr (map2 cons) s1 s2) k) l =
  if l < size s2 then nth (0:mT) (nth [::] s2 l) k else nth 0 (nth [::] s1 k) (l - size s2).
    elim=> [k l s1|t2 s2 IHs k l s1]; first by rewrite subn0.
    rewrite size_map2 => Hk; rewrite (nth_map2 0 [::]) //.
    case:l=> // l; rewrite /= IHs //.
    by rewrite leq_minr in Hk; case/andP:Hk.
  by rewrite size_seqmx ltn_ord mxE -seqmxE.
by rewrite size_trseqmx.
Qed.

End SeqmxOp.

Section SeqmxOp2.

Variables m n p' : nat.
Local Notation p := p'.+1.

Definition mulseqmx (M N: seqmatrix) : seqmatrix :=
  let N := trseqmx N in
  map (fun r => map (foldl2 (fun z x y => x * y + z) 0 r) N) M.

Lemma minSS (p q : nat) : minn p.+1 q.+1 = (minn p q).+1.
Proof. by rewrite /minn ltnS; case:ifP. Qed.

Lemma mulseqmxE (M:'M_(m,p)) (N:'M_(p,n)) :
  mulseqmx (seqmx_of_mx M) (seqmx_of_mx N) = seqmx_of_mx (M *m N).
Proof.
apply/seqmxP; split => [|i Hi|i j]; first by rewrite size_map size_seqmx.
  rewrite /rowseqmx (nth_map [::]) size_map.
    by rewrite size_trseqmx.
  by rewrite size_enum_ord.
rewrite /mulseqmx mxE /fun_of_seqmx /rowseqmx (nth_map [::]).
  rewrite (nth_map [::]); last by rewrite size_trseqmx.
  have-> : forall s1 s2 (x : mT), (foldl2 (fun z x y => x * y + z) x s1 s2) =
    x + \sum_(0 <= k < minn (size s1) (size s2)) s1`_k * s2`_k.
    elim=>[s2 x|t1 s1 IHs s2 x].
      by rewrite min0n big_mkord big_ord0 GRing.addr0.
    case:s2=>[|t2 s2]; first by rewrite minn0 big_mkord big_ord0 GRing.addr0.
    by rewrite /= IHs minSS big_nat_recl [_ + x]GRing.addrC GRing.addrA.
  rewrite GRing.add0r size_row_seqmx // size_row_trseqmx // minnn big_mkord.
  by apply:eq_bigr=>k _; rewrite trseqmxE -!fun_of_seqmxE !seqmxE mxE.
by rewrite size_seqmx.
Qed.

End SeqmxOp2.

Lemma mulseqmxEnn : forall m (M:'M_m) (N:'M_m),
  mulseqmx (seqmx_of_mx M) (seqmx_of_mx N) = seqmx_of_mx (M *m N).
Proof. by case=>[M N|m M N]; rewrite ?(flatmx0 M) ?mul0mx ?seqmx0n ?mulseqmxE. Qed.

Section SeqmxRowCol.

Variables m m1 m2 n n1 n2 : nat.

(* Block operations *)
(*get the vector whose components are in the position >= i, di dimension of i*)
Definition subseqrow i di (s : seqrow) :=
  take di (drop i s).

(*get a submatrix from the position (i, j), di and dj size of M*)
Definition sub_seqmx i j di dj (M : seqmatrix) :=
  map (subseqrow j dj) (take di (drop i M)).

(*take the m1 first row of the matrix*)
Definition usubseqmx (M : seqmatrix) :=
  take m1 M.

(*delete the m1 first rows of M*)
Definition dsubseqmx (M : seqmatrix) :=
  drop m1 M.


(*get the n1 first columns*)
Definition lsubseqmx (M : seqmatrix) :=
  map (take n1) M.

(*delete the n1 first columns*)
Definition rsubseqmx (M : seqmatrix) :=
  map (drop n1) M.




Lemma usubseqmxE : forall (M:'M[mT]_(m1+m2,n)),
  usubseqmx (seqmx_of_mx M) = seqmx_of_mx (usubmx M).
Proof.
move=> M ; apply/seqmxP ; split=> [|i Hi|i j].
  + rewrite size_take !size_seqmx; case:m2=> [|p] ; first by rewrite addn0 ltnn.
    by rewrite addnS leq_addr.
  + by rewrite /rowseqmx nth_take // size_row_seqmx //; exact:ltn_addr.
  by rewrite /fun_of_seqmx /rowseqmx nth_take // mxE -seqmxE.
Qed.

Lemma dsubseqmxE : forall (M:'M[mT]_(m1+m2,n)),
  dsubseqmx (seqmx_of_mx M) = seqmx_of_mx (dsubmx M).
Proof.
move=> M ; apply/seqmxP ; split=> [|i Hi|i j].
  + by rewrite size_drop !size_seqmx addKn.
  + by rewrite /rowseqmx nth_drop // size_row_seqmx // ltn_add2l.
  by rewrite /fun_of_seqmx /rowseqmx nth_drop // mxE -seqmxE.
Qed.

Lemma lsubseqmxE : forall (M:'M[mT]_(m,n1+n2)),
  lsubseqmx (seqmx_of_mx M) = seqmx_of_mx (lsubmx M).
Proof.
move=> M ; apply/seqmxP ; split=> [|i Hi|i j].
  + by rewrite size_map size_seqmx.
  + rewrite /rowseqmx (nth_map [::]) ?size_seqmx // size_take size_row_seqmx //.
    case:n2=> [|p] ; first by rewrite addn0 ltnn.
    by rewrite addnS leq_addr.
  + rewrite /fun_of_seqmx /rowseqmx (nth_map [::]) ?size_seqmx //.
  by rewrite nth_take // mxE -seqmxE.
Qed.

Lemma rsubseqmxE : forall (M:'M[mT]_(m,n1+n2)),
  rsubseqmx (seqmx_of_mx M) = seqmx_of_mx (rsubmx M).
Proof.
move=> M ; apply/seqmxP ; split=> [|i Hi|i j].
  + by rewrite size_map size_seqmx.
  + rewrite /rowseqmx (nth_map [::]) ?size_seqmx // size_drop.
    by rewrite size_row_seqmx // addKn.
  + rewrite /fun_of_seqmx /rowseqmx (nth_map [::]) ?size_seqmx //.
  by rewrite nth_drop // mxE -seqmxE.
Qed.

End SeqmxRowCol.

Section SeqmxRowCol2.

Variables m m1 m2 n n1 n2 : nat.

Definition ulsubseqmx (M : seqmatrix) :=
  map (take n1) (take m1 M).

Definition ursubseqmx (M : seqmatrix) :=
  map (drop n1) (take m1 M).

Definition dlsubseqmx (M : seqmatrix) :=
  map (take n1) (drop m1 M).

Definition drsubseqmx (M : seqmatrix) :=
  map (drop n1) (drop m1 M).

Lemma ulsubseqmxE (M:'M[mT]_(m1+m2,n1+n2)) :
  ulsubseqmx (seqmx_of_mx M) = seqmx_of_mx (ulsubmx M).
Proof. by rewrite -lsubseqmxE -usubseqmxE. Qed.

Lemma ursubseqmxE (M:'M[mT]_(m1+m2,n1+n2)) :
  ursubseqmx (seqmx_of_mx M) = seqmx_of_mx (ursubmx M).
Proof. by rewrite -rsubseqmxE -usubseqmxE. Qed.

Lemma dlsubseqmxE (M:'M[mT]_(m1+m2,n1+n2)) :
  dlsubseqmx (seqmx_of_mx M) = seqmx_of_mx (dlsubmx M).
Proof. by rewrite -lsubseqmxE -dsubseqmxE. Qed.

Lemma drsubseqmxE (M:'M[mT]_(m1+m2,n1+n2)) :
  drsubseqmx (seqmx_of_mx M) = seqmx_of_mx (drsubmx M).
Proof. by rewrite -rsubseqmxE -dsubseqmxE. Qed.

Definition row_seqmx (M N : seqmatrix) :=
  let aux P Q j := if j < size P then nth 0 P j else nth 0 Q (j-size P) in
  mkseq (fun i => mkseq (aux (nth [::] M i) (nth [::] N i)) (size (rowseqmx M i) + size (rowseqmx N i))) (size M).




Definition col_seqmx (M N : seqmatrix) :=
  mkseq (fun i => if i < size M then nth [::] M i else nth [::] N (i-size M)) (size M + size N).

Definition block_seqmx Aul Aur Adl Adr : seqmatrix :=
  col_seqmx (row_seqmx Aul Aur) (row_seqmx Adl Adr).

Definition block_seqmx2 Aul Aur Adl Adr : seqmatrix :=
  (map2 cat Aul Aur) ++ (map2 cat Adl Adr).

Lemma row_seqmxE (A1 : 'M_(m, n1)) (A2 : 'M_(m, n2)) :
  row_seqmx (seqmx_of_mx A1) (seqmx_of_mx A2) = seqmx_of_mx (row_mx A1 A2).
Proof.
apply/seqmxP; split=>[|i Hi|i j] ; first by rewrite size_mkseq !size_seqmx.
  by rewrite /fun_of_seqmx /rowseqmx nth_mkseq ?size_seqmx // size_mkseq !size_row_seqmx.
rewrite mxE /fun_of_seqmx /rowseqmx !nth_mkseq ?size_seqmx //.
  case:(splitP j)=> j' -> ; first by rewrite size_row_seqmx ltn_ord // -seqmxE.
  by rewrite !size_row_seqmx // ltnNge leq_addr addKn -seqmxE.
by rewrite !size_row_seqmx.
Qed.

Lemma col_seqmxE (A1 : 'M_(m1, n)) (A2 : 'M_(m2, n)) :
  col_seqmx (seqmx_of_mx A1) (seqmx_of_mx A2) = seqmx_of_mx (col_mx A1 A2).
Proof.
apply/seqmxP; split=>[|i Hi|i j] ; first by rewrite size_mkseq !size_seqmx.
  rewrite /fun_of_seqmx /rowseqmx nth_mkseq !size_seqmx //.
  case:ifPn=>H ; first exact:size_row_seqmx.
  rewrite -leqNgt in H ; have Hi' : i - m1 < m2.
    by rewrite -(addKn m1 m2) -(ltn_add2r m1) addKn (subnK H) addnC.
  by rewrite size_row_seqmx.
rewrite mxE /fun_of_seqmx /rowseqmx nth_mkseq.
  case:(splitP i)=> i' ->.
    by rewrite size_seqmx ltn_ord -seqmxE.
  by rewrite size_seqmx ltnNge leq_addr addKn -seqmxE.
by rewrite !size_seqmx.
Qed.

End SeqmxRowCol2.



Section SeqmxBlock.

Variables m1 m2 n1 n2 : nat.

Lemma block_seqmxE (Aul : 'M_(m1,n1)) (Aur:'M_(m1,n2)) (Adl:'M_(m2,n1)) (Adr:'M[mT]_(m2,n2)) :
  block_seqmx (seqmx_of_mx Aul) (seqmx_of_mx Aur)
  (seqmx_of_mx Adl) (seqmx_of_mx Adr) =
  seqmx_of_mx (block_mx Aul Aur Adl Adr).
Proof. by rewrite /block_seqmx !row_seqmxE col_seqmxE. Qed.

Lemma cast_seqmx (M:'M[mT]_(m1,n1)) (H1 : m1=m2) (H2 : n1=n2) :
  seqmx_of_mx (castmx (H1,H2) M) = seqmx_of_mx M.
Proof.
apply/seqmxP; split=> [|i Hi| i j] ; first by rewrite size_seqmx.
  by rewrite size_row_seqmx ?H2 // -H1.
rewrite -seqmxE ; congr fun_of_seqmx ; move: (H1) (H2) M ; rewrite H1 H2=> H3 H4 M.
by rewrite castmx_id.
Qed.

Definition xrowseqmx (M : seqmatrix) : seqmatrix :=
  let R := set_nth [::] M m1 (nth [::] M m2) in
  set_nth [::] R m2 (nth [::] M m1).

End SeqmxBlock.

Definition scalar_seqmx (n : nat) (x : mT) :=
  mkseqmx n n (fun i j => x *+ (i == j)).

Lemma scalar_seqmxE n x :
  scalar_seqmx n x = seqmx_of_mx (@scalar_mx mT n x).
Proof.
apply/seqmxP ; split=> [||i j] ; first by rewrite size_map size_iota.
  exact:size_row_mkseqmx.
by rewrite mkseqmxE // mxE.
Qed.

Definition scaleseqmx (x : mT) (M : seqmatrix) :=
  map (map ( *%R x)) M.

Lemma scaleseqmxE m n x (M : 'M_(m,n)) :
  scaleseqmx x (seqmx_of_mx M) = seqmx_of_mx (scalemx x M).
Proof.
apply/seqmxP; split=> [|i Hi| i j] ; first by rewrite size_map size_seqmx.
  by rewrite /rowseqmx (nth_map [::]) ?size_seqmx // size_map size_row_seqmx.
rewrite mxE /fun_of_seqmx /rowseqmx (nth_map [::]) ?size_seqmx // (nth_map 0).
  by rewrite -seqmxE.
by rewrite size_row_seqmx.
Qed.

Definition const_seqmx m n (x : mT) := nseq m (nseq n x).

Lemma const_seqmxE m n x : const_seqmx m n x = @seqmx_of_mx m n (const_mx x).
Proof.
apply/seqmxP; split=> [|i Hi|i j]; first by rewrite size_nseq.
  by rewrite /rowseqmx nth_nseq Hi size_nseq.
by rewrite /fun_of_seqmx /rowseqmx nth_nseq (ltn_ord i) nth_nseq (ltn_ord j) mxE.
Qed.

Definition seqmx0 m n := const_seqmx m n 0.

Lemma seqmx0E m n : seqmx_of_mx (0: 'M[mT]_(m,n)) = seqmx0 m n.
Proof. by rewrite /seqmx0 -const_seqmxE. Qed.

Definition seqmx1 n := scalar_seqmx n 1.

Lemma seqmx1E n : seqmx_of_mx (1%:M : 'M[mT]_n) = seqmx1 n.
Proof. by rewrite /seqmx1 -scalar_seqmxE. Qed.


End seqmx.
