Add LoadPath "CoqEAL".

Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq choice fintype.
Require Import div finfun bigop prime binomial ssralg finset fingroup finalg.
Require Import perm zmodp matrix mxalgebra seqmatrix1 triangular matrix_facts permseq.

(*****************************************************************************)
(*  mx_of_seqmx_seqmx_of_mx == cancel mx_of_seqmx seqmx_of_mx                *)
(*  seqmx_of_mx_mx_of_seqmx == cancel seqmx_of_mx mx_of_seqmx                *)
(*  det_seqmx == determinant of executable matrices
      (developed by Y. Bertot and M. Dénès) *)
(*  invmx_seqmx == inverse of executable matrices *)
(*  rank == rank over executable matrices
      (developed by M. Dénès, A. Mörtberg and V. Siles) *)
(*  lower_triangular_seq == checks that an executable matrix is a lower
      triangular matrix with 1's in the diagonal *)
(*****************************************************************************)


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


Local Open Scope ring_scope.



Section identity_lemmas.

(*
We prove the equivalence, module a domain transformation, of the types seqMatrix and SSReflect matrix
*)

Variable mT : ringType.


Variables m n : nat.

Lemma mxseqE: forall (M: seqmatrix mT) (i:'I_m)(j:'I_n),
M i j = (mx_of_seqmx m n M) i j.
Proof.
by move => M i j; rewrite mxE.
Qed.

Lemma mxseqP (M : seqmatrix mT) (N : 'M_(m,n)) :
  [/\ m = size M, forall i, i < m -> size (rowseqmx M i) = n &
  mx_of_seqmx m n M = N] <-> M = seqmx_of_mx N.
Proof.
split.
move => [Hm Hn HMN].
rewrite -seqmxP.
split => //=.
by rewrite -HMN => i j; rewrite mxseqE.
rewrite -seqmxP.
case => H; split =>//=.
apply /matrixP => i j.
rewrite - p0.
by rewrite (mxseqE M).
Qed.


Lemma mx_of_seqmx_seqmx_of_mx: forall (M: 'M[mT]_(m,n)),
  mx_of_seqmx m n (seqmx_of_mx (m:=m) (n:=n)  M) = M.
Proof.
by move => M; by apply mxseqP.
Qed.


Lemma seqmx_of_mx_mx_of_seqmx: forall (M : seqmatrix mT),
  [/\ m = size M & forall i, i < m -> size (rowseqmx M i) = n]->
  seqmx_of_mx (mx_of_seqmx m n M) = M.
Proof.
move => M; case => H H1.
have :  M = seqmx_of_mx (mx_of_seqmx m n M).
apply mxseqP. split => //=.
by move => H2.
Qed.


End identity_lemmas.

Section opposite.

Variable mT : ringType.

Definition oppseqmx (M : seqmatrix mT) : seqmatrix mT :=
  map_seqmx (fun x => GRing.opp x) M.

Variables m n:nat.

Lemma oppseqmxE:
  {morph (@seqmx_of_mx mT m n) : M / - M >-> oppseqmx M}.
Proof. by rewrite /oppseqmx => M; rewrite -map_seqmxE. Qed.

End opposite.


Section mulseqmx_lemmas.

(* The multiplication of objects of type seqmatrix is associative *)

Variable mT : ringType.

Lemma mulseqmxA (X Y Z: seqmatrix mT) (x1 x2 y z:nat):
[/\ x1.+1 = size X & forall i, i < x1.+1 -> size (rowseqmx X i) = x2.+1] ->
[/\ x2.+1 = size Y & forall i, i < x2.+1 -> size (rowseqmx Y i) = y.+1] ->
[/\ y.+1 = size Z & forall i, i < y.+1 -> size (rowseqmx Z i) = z.+1] ->
mulseqmx X (mulseqmx Y Z) = mulseqmx (mulseqmx X Y) Z.
Proof.
move=> H1 H2 H3.
by rewrite -(seqmx_of_mx_mx_of_seqmx (m:=x1.+1) (n:=x2.+1) (M:= X)) //=
           -(seqmx_of_mx_mx_of_seqmx (m:=x2.+1) (n:=y.+1) (M:= Y)) //=
           -(seqmx_of_mx_mx_of_seqmx (m:=y.+1) (n:=z.+1) (M:= Z)) //=
           -(seqmx_of_mx_mx_of_seqmx (m:=x1.+1) (n:=x2.+1) (M:= X)) //=
           -(seqmx_of_mx_mx_of_seqmx (m:=x2.+1) (n:=y.+1) (M:= Y)) //=
           -(seqmx_of_mx_mx_of_seqmx (m:=y.+1) (n:=z.+1) (M:= Z)) //=
           !mulseqmxE mulmxA.
Qed.

End  mulseqmx_lemmas.



Section determinant.

Variable R : comRingType.

Fixpoint expand sign (mat : seqmatrix R) det acc : R :=
  if mat is a :: tl then
    sign * head 0 a * det (rev acc ++ (map behead tl)) +
    expand (- sign) tl det (behead a :: acc)
  else
    0.

Fixpoint det_seqmx n (mat : seqmatrix R) : R :=
  if n is p.+1 then expand 1 mat (det_seqmx p) [::] else 1.

Lemma can_seqmx_of_mx : forall n m (mat:'M[R]_(n,m)), mx_of_seqmx n m (seqmx_of_mx mat) = mat.
Proof.
move =>  n m mat; apply/matrixP => x y.
rewrite /mx_of_seqmx /seqmx_of_mx /rowseqmx mxE.
rewrite (nth_map x); last by rewrite size_enum_ord ltn_ord.
rewrite (nth_map y); last by rewrite size_enum_ord ltn_ord.
by rewrite !nth_ord_enum.
Qed.

Lemma castmx_det :
  forall n p (m : 'M[R]_n) (h : n = p), \det m = \det (castmx (h, h) m).
Proof.
 by move => n p m h; move: (h) m; rewrite h => h' m; rewrite castmx_id.
Qed.

Lemma col'0_mx_of_seqmx_behead :
  forall l p mat, col' ord0 (@mx_of_seqmx R l p.+1 mat) =
       mx_of_seqmx l p (map behead mat).
Proof.
move => l p mat; rewrite /col' /mx_of_seqmx /rowseqmx; apply/matrixP => i j; rewrite !mxE /=.
case w: (i < size mat).
 rewrite (nth_map ([::]:seq R)) // /behead.
 by  case: (nth _ _ i); case j => [ [| ?] ?].
by rewrite !nth_default // ?size_map leqNgt negbT.
Qed.

Lemma col'0_mx_of_seqmx :
  forall l1 p n mat (h : p.+1 = n),
    col' (cast_ord h ord0) (@mx_of_seqmx R l1 n mat) =
    mx_of_seqmx l1 n.-1 (map behead mat).
Proof.
by move => l1 p n mat h; case h; rewrite cast_ord_id col'0_mx_of_seqmx_behead.
Qed.

Lemma cast_mx_of_seqmx :
  forall n1 n2 n3 n4 (q : n1 = n2) (h : n3 = n4) m,
    castmx (q, h) (@mx_of_seqmx R n1 n3 m) = mx_of_seqmx n2 n4 m.
Proof.
by move => n1 n2 n3 n4 q h; move: (q) (h); rewrite q h => q' h' m;
   rewrite castmx_id.
Qed.

Lemma mx_of_seqmx_cat :
  forall k l p m1 m2, size m1 = k ->
    (@mx_of_seqmx R (k + l) p (m1 ++ m2)) = col_mx (mx_of_seqmx k p m1) (mx_of_seqmx l p m2).
Proof.
move => k l p m1 m2 s1; apply/matrixP => i j; rewrite /mx_of_seqmx /col_mx /rowseqmx !mxE.
rewrite nth_cat s1; case: splitP => [i' -> | j' ->]; rewrite mxE //.
by rewrite -{2}[k]addn0 subn_add2l subn0.
Qed.

Lemma zip_rcons : forall T1 T2 (l1:seq T1)(l2:seq T2) a b,
  size l1 = size l2 -> zip (rcons l1 a) (rcons l2 b) = rcons (zip l1 l2) (a,b).
Proof.
move => T1 T2; elim => [ | c l1 Hl1]; first by case.
by move => [ | d l2] a b //; move => [q] /=; rewrite Hl1.
Qed.

Lemma rev_zip : forall T1 T2 (l1:seq T1)(l2:seq T2),
   size l1 = size l2 -> rev (zip l1 l2) = zip (rev l1) (rev l2).
Proof.
move => T1 T2; elim => [|a l1 Hl1] [|b l2] // [sl].
by rewrite !rev_cons zip_rcons ?size_rev // Hl1.
Qed.

Lemma row'0_mx_of_seqmx : forall m1 r m2 k l p, size m1 = k ->
   row' (rshift k 0) (@mx_of_seqmx R (k + l.+1) p (m1 ++ (cons r m2))) =
   mx_of_seqmx (k + l.+1).-1 p (m1 ++ m2).
Proof.
move => m1 r m2 k l p sm1; apply/matrixP => i j; rewrite !mxE /= /bump addn0.
case q: (k <= i).
 by rewrite /rowseqmx !nth_cat add1n sm1 !ltnNge leqW //= leq_subS q.
by rewrite /rowseqmx !nth_cat add0n ltnNge sm1 q.
Qed.

Definition czip (v:seq R) (l: seq (seq R)) :=
  map (fun p => let: (x,y) := p in x::y) (zip v l).

Lemma expand_correct :
  forall mat p (k l:nat) (f : seq (seq R) -> R)
   s acc v (h :(p.+1 = k + l)%N),
   size acc = k -> size mat <= l -> size v = k ->
    (forall mat', size mat' <= p -> f mat' = \det (mx_of_seqmx p p mat')) ->
    let r := mx_of_seqmx (k+l) p.+1 (rev (czip v acc) ++ mat) in
    expand s mat f acc =
      s * (-1)^+ k * \sum_(i < l) r (rshift k i) 0 *
      cofactor (castmx (refl_equal (k + l)%N, h) r)
         (rshift k i) (cast_ord h ord0).
Proof.
elim => [ | r mat' Ih] p k l f s acc v h sa sm sv fc.
 set ref := mx_of_seqmx _ _ _; rewrite /= big1 ?GRing.mulr0 // => i _.
 suffices -> :  ref (rshift k i) 0 = 0 by rewrite GRing.mul0r.
 rewrite /ref mx_of_seqmx_cat; first by rewrite col_mxEd mxE !nth_default /rowseqmx ?nth_nil //=.
 by rewrite size_rev /czip size_map size_zip sa sv minnn.
move: h sm; case ld : l => [ | l'] // h /= sm; set ref := mx_of_seqmx _ _ _.
have qp : p.+1 = (k.+1 + l')%N by rewrite h addnS.
have qp' : p = (k + l'.+1).-1 by rewrite addnS -addSn -qp.
have sa' : size ((behead r)::acc) = k.+1 by rewrite /= sa.
have sv' : size ((head 0 r)::v) = k.+1 by rewrite /= sv.
have sz : size (rev (czip v acc)) = k.
 by rewrite size_rev /czip size_map size_zip sv sa minnn.
move: (Ih p _ l' f (-s) _ _ qp sa' sm  sv' fc) => -> {Ih}.
rewrite big_ord_recl {1}/ref.
have -> : mx_of_seqmx (k + l'.+1) p.+1 (rev (czip v acc) ++ (r::mat'))
              (rshift k ord0) 0 = head 0 r.
 by rewrite /mx_of_seqmx /rowseqmx mxE /= nth_cat addn0 sz ltnn subnn /=.
have -> : cofactor (castmx (Logic.eq_refl (k + l'.+1)%N, h) ref)
      (rshift k ord0) (cast_ord h ord0) =
        (-1)^+k * f (rev acc ++ map behead mat').
 rewrite /cofactor /ref /= !addn0; congr (_ * _); rewrite fc; last first.
  by rewrite size_cat size_rev size_map sa qp' addnS /= leq_add2l -ltnS.
 rewrite cast_mx_of_seqmx col'0_mx_of_seqmx map_cat map_rev /czip -map_comp.
 have unzs : {in (zip v acc), funcomp tt behead
                       (fun p0 : R * seq R => let '(x, y) := p0 in cons x y) =1
                 @snd  R _} by move => [a b] _ /=.
 rewrite (eq_in_map unzs) -/unzip2 unzip2_zip; last by rewrite sa sv leqnn.
 rewrite /= row'0_mx_of_seqmx; last by rewrite size_rev.
 by rewrite -(cast_mx_of_seqmx (esym qp') (esym qp')) -castmx_det.
rewrite GRing.mulr_addr; congr (_ + _).
 rewrite !GRing.mulrA; congr (_ * _); rewrite -!GRing.mulrA; congr (_ * _).
 by rewrite GRing.mulrC -GRing.mulrA -GRing.exprn_addr -GRing.signr_odd addnn odd_double GRing.expr0 GRing.mulr1.
rewrite -GRing.mulN1r GRing.exprS -(GRing.mulrA (-1)) !(GRing.mulrC (-1)) !GRing.mulrA
 -(GRing.mulrA _ (-1) (-1)) GRing.mulrNN !GRing.mulr1 (GRing.mulrC s); congr (_ * _).
apply eq_big => // i _;rewrite !mxE /ref.
have -> : rshift k (lift ord0 i) = (k + i.+1)%N :> nat by [].
have -> : rshift k.+1 i = (k.+1 + i)%N :> nat by [].
congr (_ * _).
 by rewrite /rowseqmx !nth_cat !ltnNge !size_rev /czip !size_map !size_zip sa' sv'
   sa sv !minnn leq_addr addnS addKn (leqW (leq_addr i k)) -(addnS k i) addKn.
congr (_ * _); first by rewrite /= !addnS.
rewrite !cast_mx_of_seqmx !col'0_mx_of_seqmx.
have h' : (k + l'.+1).-1 = (k.+1 + l').-1 by rewrite addnS.
rewrite (castmx_det _ h'); apply: f_equal.
apply/matrixP => x y; rewrite /row' castmxE !mxE /czip -!map_rev.
have sc : size (cons _ v) = size (cons _ acc) by move => *; rewrite /= sv.
have sav : size v = size acc by rewrite sv.
rewrite rev_zip // !rev_cons zip_rcons ?map_rcons ?size_rev //.
by rewrite cat_rcons !map_cat !map_cons rev_zip //= /bump leq0n addnS.
Qed.

Lemma det_seqmx_correct :
  forall n mat, size mat <= n -> det_seqmx n mat = \det (mx_of_seqmx n n mat).
Proof.
elim => [ | n Ih] mat sm /=.
 by rewrite /det_seqmx /mx_of_seqmx (flatmx0 (\matrix_(i,j) _)) -(flatmx0 1%:M) det1.
rewrite (expand_correct _ (erefl _ : n.+1 = (0 + n.+1)%N)
           (erefl (size nil)) sm (erefl (size nil)) Ih).
rewrite !castmx_id (expand_det_col _ 0) GRing.expr0 !GRing.mul1r; apply: eq_bigr => i _.
by rewrite cast_ord_id (_ : rshift 0 i = i) //; apply: val_inj.
Qed.

Lemma det_seqmx_compute :
  forall n (m:'M[R]_n), \det m = det_seqmx n (seqmx_of_mx m).
Proof.
move => n m; rewrite det_seqmx_correct ?(can_seqmx_of_mx m) //.
 by rewrite /seqmx_of_mx size_map size_enum_ord.
Qed.

End determinant.



Section invmxseq.

Variable R : comUnitRingType.

Section definitions.

Definition splice_out i (A : seq R) := (take i A) ++ drop (i+1) A.

(* A with the i'th row spliced out.*)
Definition row'_seqmx i (A:seqmatrix R) :seqmatrix R := (take i A) ++ drop (i+1) A.

(* A with the j'th column spliced out.*)
Definition col'_seqmx j (A:seqmatrix R) :seqmatrix R := map (splice_out j) A.

(* the i, j cofactor of A (the signed i, j minor of A *)
Definition cofactor_seqmx n (A:seqmatrix R) i j :=
  (-1) ^+ (i + j) * (det_seqmx n.-1 (row'_seqmx i (col'_seqmx j A))).


(* The adjugate matrix : defined as the transpose of the matrix of cofactors *)

Definition adjugate_seqmx n (A:seqmatrix R) :=
  map (fun i => map (fun j => (cofactor_seqmx n A j i)) (iota 0 n)) (iota 0 n).


Definition invmx_seqmx n (A:seqmatrix R) :=
  if (GRing.unit (det_seqmx n A)) then (scaleseqmx (det_seqmx n A)^-1 (adjugate_seqmx n A)) else A.

End definitions.

Section properties.

Require Import Bool.

Lemma row'_seqmxE m n (i0:'I_m) (M:'M[R]_(m,n)) :
  row'_seqmx i0 (seqmx_of_mx M) = seqmx_of_mx (row' i0 M).
Proof.
apply/seqmxP; split=> [|i Hi|i j].
  + rewrite size_cat size_drop size_take size_seqmx (ltn_ord i0) addn_subA addn1;
    last by move : (ltn_ord i0).
    by rewrite -addn1 subn_add2l subn1.
  + rewrite /rowseqmx /row'_seqmx nth_cat.
    case H : (i < size (take i0 (seqmx_of_mx M))).
      - rewrite nth_take; last by move : H; rewrite size_take size_seqmx (ltn_ord i0) => ->.
          rewrite size_row_seqmx //; move : (ltnSn m.-1); rewrite {2}(ltn_predK (ltn_ord i0)) => H2; apply : (ltn_trans Hi H2).
       rewrite nth_drop size_row_seqmx // size_take size_seqmx (ltn_ord i0) (addnC i0 1%N) -addnA.
       rewrite subnKC;
           last by move: H; rewrite size_take size_seqmx (ltn_ord i0) => /negP; rewrite leqNgt=> /negP /negbNE.
        rewrite addnC.
        have H3 : i.+1<m.-1.+1; first by done.
        by rewrite (ltn_predK (ltn_ord i0)) in H3; rewrite addn1.
  + rewrite /fun_of_seqmx /rowseqmx nth_cat.
    case H : (i < size (take i0 (seqmx_of_mx M))).
      - rewrite nth_take; last by move : H; rewrite size_take size_seqmx (ltn_ord i0) => ->.
        rewrite mxE -seqmxE //= /bump.
       by move : H; rewrite size_take size_seqmx (ltn_ord i0) ltnNge negb_true_iff => ->.
       rewrite nth_drop // mxE -seqmxE size_take size_seqmx (ltn_ord i0) /= /bump
          (addnC i0 1%N) -addnA subnKC; last by move : H; rewrite size_take size_seqmx (ltn_ord i0) leqNgt negb_false_iff ltnS.
         by move : H; rewrite size_take size_seqmx (ltn_ord i0) leqNgt negb_false_iff ltnS => ->. 
Qed.


Lemma col'_seqmxE m n (j0:'I_n) (M:'M[R]_(m,n)) :
  col'_seqmx j0 (seqmx_of_mx M) = seqmx_of_mx (col' j0 M).
Proof.
apply/seqmxP; split=> [|i Hi|i j].
  + by rewrite size_map size_seqmx.
  + rewrite /rowseqmx (nth_map [::]); last by rewrite size_seqmx.
    by rewrite /splice_out size_cat size_drop size_take size_row_seqmx // (ltn_ord j0)
              addn_subA ?subn_add2l ?subn1 // ?addn1 ?(ltn_ord j0).
  + rewrite /fun_of_seqmx /rowseqmx (nth_map [::]) ?size_seqmx // /splice_out.
    rewrite nth_cat size_take size_row_seqmx ?(ltn_ord j0) ?(ltn_ord i) //.
    case H : (j<j0).
        rewrite mxE -seqmxE /= /bump.
          move : H; rewrite ltnNge negb_true_iff; move => H; rewrite H /= add0n. 
          rewrite nth_take //. 
         by move : H; rewrite leqNgt negb_false_iff => ->. 
        rewrite mxE -seqmxE /= /bump.
          move : H; rewrite ltnNge negb_false_iff; move => H; rewrite H /=.
          by rewrite nth_drop (addnC j0 1%N) -addnA subnKC.

Qed.

Lemma cofactor_seqmxE n (M:'M[R]_n) (i j:'I_n) : cofactor_seqmx n (seqmx_of_mx M) i j = cofactor M i j.
Proof.
by rewrite /cofactor_seqmx /cofactor col'_seqmxE row'_seqmxE det_seqmx_compute.
Qed.


Lemma adjugate_seqmxE n (M:'M[R]_n) :
   adjugate_seqmx n (seqmx_of_mx M) = (seqmx_of_mx (\adj M)).
Proof.
apply/seqmxP; split=> [|i Hi|i j].
  + by rewrite size_map size_iota.
  + rewrite /rowseqmx (nth_map 0%N); last by rewrite size_iota Hi.
    by rewrite size_map size_iota.
  + rewrite /fun_of_seqmx /rowseqmx (nth_map 0%N); last by rewrite size_iota (ltn_ord i).
    rewrite nth_iota; last by rewrite (ltn_ord i).
    rewrite add0n (nth_map 0%N); last by rewrite size_iota (ltn_ord j).
    rewrite nth_iota; last by rewrite (ltn_ord j).
    by rewrite add0n cofactor_seqmxE /adjugate mxE.
Qed.

Lemma invmx_seqmxE n (M:'M[R]_n) :
   invmx_seqmx n (seqmx_of_mx M) = (seqmx_of_mx (invmx M)).
Proof.
rewrite /invmx_seqmx.
case H : (GRing.unit (det_seqmx n (seqmx_of_mx M))).
  + rewrite adjugate_seqmxE scaleseqmxE -det_seqmx_compute /invmx.
    by have -> : M \in unitmx; first by rewrite unitmxE det_seqmx_compute H.
rewrite /invmx.
have -> : (M \in unitmx)=false; last by done.
move : H.
by rewrite -det_seqmx_compute -unitmxE.
Qed.

Lemma invmx_seqmx2E n (M:seqmatrix R) :
[/\ n = size M & forall i, i < n -> size (rowseqmx M i) = n] ->
   invmx (mx_of_seqmx n n M) = (mx_of_seqmx n n (invmx_seqmx n M)).
Proof.
move => H.
by rewrite -{2}(seqmx_of_mx_mx_of_seqmx H) invmx_seqmxE mx_of_seqmx_seqmx_of_mx.
Qed.





End properties.

End invmxseq.



Section FieldRank.

Import GRing.Theory.

Variable K : fieldType.

Local Open Scope ring_scope.



Lemma rank0M': forall (n m:nat) (M: 'M[K]_(n,m)),
  \rank (row_mx (0: 'cV[K]_n) M) = \rank M.
Proof.
move => n m M.
have -> : row_mx 0 M = (col_mx 0 M^T)^T.
- by move => a; rewrite -[X in (col_mx X _)]trmx0 tr_col_mx !trmxK.
by rewrite mxrank_tr eqmxM0M mxrank_tr.
Qed.

Lemma tool_rank : forall n:nat, n != 0%N -> n <= 1 -> n = 1%N.
Proof. by case => [ | [ | n]]. Qed.

Lemma inv_scalar0 : forall n (a:K), a%:M = 0 :> 'M[K]_n.+1 -> a = 0.
Proof.
move => n a.
rewrite -scalemx1 => /matrixP  h.
by move: (h 0 0); rewrite !mxE mulr1.
Qed.

Lemma rank0M: forall (n m:nat) (M: 'M[K]_(n,m)),
  \rank (col_mx (0: 'rV[K]_m) M) = \rank M.
Proof. by move => n m M; rewrite eqmxM0M. Qed.

(* rank ( a    0 .... 0  )  = 1 + rank M if a != 0
          c1
          ...     M
          cn
*)
Lemma rankaM : forall (n m:nat) (M: 'M[K]_(n,m)) (a:K) (c:'cV[K]_n),
 a != 0 -> \rank (block_mx (a%:M) 0 c M) = (1 + \rank M)%N.
Proof.
move => n m M a c ha.
have: ((block_mx a%:M 0 c M)^T :=: (col_mx a%:M c)^T + (col_mx 0 M)^T)%MS.
- rewrite tr_block_mx !tr_col_mx /block_mx.
  apply eqmx_sym.
  by apply addsmxE.
have hC: (col_mx a%:M c) != 0.
- apply/negP; move/eqP => h.
  move/negP : ha; apply.
  move/colP : h => h.
  move: (h 0); rewrite !mxE.
  case: splitP => i hi //.
  by rewrite [i]ord1 !mxE /= mulr1n => ->.
rewrite -mxrank_tr => ->.
rewrite mxrank_disjoint_sum.
- rewrite rank_rV mxrank_tr rank0M.
  by rewrite -trmx_neq0 hC.
rewrite !tr_col_mx !tr_scalar_mx trmx0.
apply/eqP.
apply/rowV0P => v.
rewrite sub_capmx; case/andP.
case/sub_rVP => b hb.
case/submxP => B hB.
have : v 0 0 = 0 /\ v 0 0 = b * a.
- split.
  + rewrite hB !mxE big1 // => i _.
    rewrite !mxE.
    case: splitP => j //.
    by rewrite !mxE /= mulr0.
  rewrite hb !mxE.
  case: splitP => j //.
  by rewrite [j]ord1 !mxE /= mulr1n.
rewrite {3}hb; case => ->.
rewrite -[0](@mul0r _ a). move/(mulIf ha) <-.
by rewrite scale0r.
Qed.

Lemma rankaM' : forall (n m:nat) (M: 'M[K]_(n,m)) (a:K) (l:'rV[K]_m),
 a != 0 -> \rank (block_mx (a%:M) l 0 M) = (1 + \rank M)%N.
Proof.
move => n m M a l ha.
rewrite -mxrank_tr tr_block_mx tr_scalar_mx trmx0 rankaM //.
by rewrite mxrank_tr.
Qed.

Lemma tcast: forall p m:nat, (p + (1 + m))%N = (p.+1 + m)%N.
Proof.
by move => p m; rewrite addnA addn1.
Qed.

Fixpoint ssr_tool (n m p:nat) (vs: 'M[K]_(m,n)) {struct p}:
  'M[K]_(p, 1 + n) ->  ('M[K]_(p + m,n) * bool) :=
   match p return 'M[K]_(p, 1 + n) -> ('M[K]_(p + m, n) * bool) with
   | p'.+1 => fun (l : 'M_(1 + p', 1 + n)) =>
     let a := l 0 0 in
     if a == 0 then
       let: (R,b) := ssr_tool (col_mx (ursubmx l) vs) (dsubmx l)
         in (castmx (tcast p' m,refl_equal n) R,b)
     else
       let v := a^-1 *: dlsubmx l in
       let R := drsubmx l - v *m ursubmx l in
       let v0 : 'rV[K]_n := 0 in
         (castmx (tcast p' m,refl_equal n) (col_mx R (col_mx v0 vs)),
           true)
   | _ => fun l => (vs, false)
end.

Fixpoint ssr_rank (m n:nat) {struct n} : 'M[K]_(m,n) -> nat :=
  match n return 'M[K]_(m,n) -> nat with
   |  q.+1 => fun M => let:(R,b) := @ssr_tool _ 0%N _ 0 M in
     (b + ssr_rank R)%N
   | _ => fun _ => 0%N
end.


Lemma ssr_tool_col0 : forall p m n (vs: 'M[K]_(m,n)) (M: 'M[K]_(p, 1 + n)),
 let: (_,b) := ssr_tool vs M in
 if b then True else lsubmx M = 0.
Proof.
Admitted.

Lemma ssr_toolP : forall p m n (vs: 'M[K]_(m,n)) (M: 'M[K]_(p,1 + n)),
  let: (R,b) := ssr_tool vs M in
    if b then \rank ((row_mx 0 vs) + M)%MS = (1 + \rank R)%N
      else (R :=: vs + rsubmx M)%MS.
Proof.
elim => [ m n vs M | p hi] /=.
- rewrite [rsubmx M]flatmx0.
  apply eqmx_sym.
  by apply addsmx0.
rewrite [p.+1]/(1 + p)%N => m n vs M /=.
case: ifP => [hM00 | ].
case: ssr_tool (hi _ _ (col_mx (ursubmx M) vs) (dsubmx M)) => R b.
- case: ifP => hb hR.
  + rewrite eqmx_cast -hR.
    have -> :
      row_mx 0 (col_mx (ursubmx M) vs) = col_mx (usubmx M) (row_mx 0 vs).
    * rewrite -[usubmx M]hsubmxK [lsubmx (usubmx M)]mx11_scalar !mxE
        !lshift0 (eqP hM00) scalar0.
      by rewrite [X in _ = X]block_mxEh col_mx0.
    transitivity (\rank (row_mx 0 vs + (usubmx M + dsubmx M))).
    * apply adds_eqmx; first by apply eqmx_refl.
      rewrite -{1}[M]vsubmxK; apply eqmx_sym.
      by apply addsmxE.
    rewrite addsmxA.
    apply adds_eqmx; last by apply eqmx_refl.
    rewrite addsmxC.
    by apply addsmxE.
  apply (eqmx_trans (eqmx_cast _ _)).
  apply (eqmx_trans hR).
  apply (eqmx_trans (adds_eqmx (eqmx_sym (addsmxE (ursubmx M) vs))
                               (eqmx_refl _))).
  rewrite [(_ + vs)%MS]addsmxC -addsmxA.
  apply adds_eqmx; first by apply eqmx_refl.
  apply (eqmx_trans (addsmxE (ursubmx M) (drsubmx M))).
  rewrite -{3}[M]vsubmxK -[usubmx M]hsubmxK -[dsubmx M]hsubmxK.
  rewrite [col_mx (row_mx _ _) (row_mx _ _)]block_mxEh row_mxKr.
  by apply: eqmx_refl.
rewrite eqmx_cast.
move/negbT => hM00.
rewrite addsmxE.
set D : 'M[K]_(1 + p) :=
  block_mx 1%:M 0 (-1 / M 0 0 *: (dlsubmx M)) 1%:M.
have hD : row_full (block_mx (1%:M : 'M[K]_m) 0 0 D).
- by rewrite row_full_unit unitmxE !det_lblock !det1 !mul1r unitr1.
rewrite -(eqmxMfull (col_mx (row_mx 0 vs) M) hD) -{1}[M]submxK
  [ulsubmx M]mx11_scalar !mxE !lshift0 block_mxEh mul_row_col.
rewrite !mul_col_mx !mul1mx !mul0mx.
rewrite [row_mx 1%:M 0 *m block_mx _ _ _ _]mul_row_col mul1mx mul0mx addr0.
rewrite [row_mx ((-1 / M 0 0) *: dlsubmx M) _ *m _]mul_row_col mul1mx
 mul_mx_row add_row_mx mul_mx_scalar scalerA mulrC !mulNr -mulrA mulVr
 ?unitfE // !mul1r scaleNr scale1r addNr add_col_mx add0r addr0.
rewrite [drsubmx M - _]addrC.
rewrite scaleNr mulNmx.
set Mdr := - ((M 0 0)^-1 *: _ *m _) + _.
transitivity (\rank
  (col_mx (row_mx (col_mx (M 0 0)%:M 0) (col_mx (ursubmx M) Mdr))
                 (row_mx 0 vs))).
- by rewrite -!addsmxE addsmxC -block_mxEh.
transitivity (\rank (block_mx (M 0 0)%:M (ursubmx M) 0 (col_mx Mdr vs))).
- rewrite -!addsmxE -block_mxEh (adds_eqmx (eqmx_sym (addsmxE _ _))
   (eqmx_refl _)).
  rewrite -[X in row_mx X (col_mx _ _)]col_mx0 -block_mxEh -addsmxA.
  apply adds_eqmx; first by apply eqmx_refl.
  by apply addsmxE.
transitivity (1 + \rank (col_mx Mdr vs))%N.
- by rewrite rankaM' // -addsmxE addsmxC addsmxE.
congr (_ + _)%N.
rewrite -!addsmxE.
apply adds_eqmx; first by apply eqmx_refl.
apply (eqmx_trans (eqmx_sym (@addsmx0 _ _ 1 _ vs))).
rewrite addsmxC.
by apply addsmxE.
Qed.

Lemma ssr_rankP : forall n m (M: 'M[K]_(m,n)),
  ssr_rank M = \rank M.
Proof.
elim => [ | n hi].
- by move => m M; rewrite [M]thinmx0 mxrank0.
rewrite [n.+1]/(1 + n)%N => m M /=.
move: (@ssr_toolP _ 0%N _ 0 M).
move: (@ssr_tool_col0 _ 0%N _ 0 M).
case: ssr_tool => R b.
case: ifP => hb => [ _ | ].
- by rewrite row_mx0 adds0mx hi => ->.
rewrite -{3}[M]hsubmxK => ->.
rewrite rank0M' hi => ->.
by rewrite adds0mx.
Qed.

Fixpoint tool p (vs : seqmatrix K) {struct p}:
  seqmatrix K ->  (seqmatrix K * bool) :=
   match p return seqmatrix K -> (seqmatrix K * bool) with
   | p'.+1 => fun l =>
     let a := l 0%N 0%N in
     if a == 0 then
       tool p' (col_seqmx (ursubseqmx 1 1 l) vs) (dsubseqmx 1 l)
     else
       let v := scaleseqmx a^-1 (dlsubseqmx 1 1 l) in
       let R := subseqmx (drsubseqmx 1 1 l) (mulseqmx v (ursubseqmx 1 1 l)) in
       let v0 := mkseqmx 1 (size (rowseqmx l 0)).-1 (fun _ _ => 0) in
         (col_seqmx R (col_seqmx v0 vs), true)
   | _ => fun l => (vs, false)
end.



Lemma toolE : forall p m n (vs : 'M_(m, n)) (M : 'M_(p, 1 + n)),
  let (R,b) := (ssr_tool vs M) in
  tool p (seqmx_of_mx vs) (seqmx_of_mx M) = (seqmx_of_mx R,b).
Proof.
elim=> // p IHs m n vs.
rewrite -add1n=> M.
rewrite /ssr_tool /= -seqmxE.
case:ifP=> //= H.
  move:(IHs _ _ (col_mx (ursubmx M) vs) (dsubmx M)).
  rewrite -/(ssr_tool _ _); case:(ssr_tool _ _)=> R b H1.
  by rewrite ursubseqmxE col_seqmxE dsubseqmxE H1 cast_seqmx.
set mk0seqmx := mkseqmx 1 _ _.
have -> : mk0seqmx = seqmx_of_mx (0 : 'rV_n).
  apply/seqmxP; split=> [|i Hi|i j].
    + by rewrite size_mkseq.
    + by rewrite size_row_mkseqmx // (@size_row_seqmx _ _ _ M 0).
    + by rewrite mkseqmxE ? mxE // (@size_row_seqmx _ _ _ M 0).
rewrite cast_seqmx -!col_seqmxE subseqmxE -drsubseqmxE -mulseqmxE -scaleseqmxE.
by rewrite -ursubseqmxE -dlsubseqmxE.
Qed.

Fixpoint rank (m n:nat) {struct n} : seqmatrix K -> nat :=
  match n return seqmatrix K -> nat with
   |  q.+1 => fun M => let:(R,b) := @tool m [::] M in
     (b + rank m q R)%N
   | _ => fun _ => 0%N
end.

Lemma rankE : forall m n (M : 'M_(m, n)),
  rank m n (seqmx_of_mx M) = ssr_rank M.
Proof.
move=> m; elim=> // n IHn M.
rewrite /rank /ssr_rank; move:(@toolE m 0 n 0 M); case:(ssr_tool _ _).
by rewrite addn0=> R b; rewrite seqmx0n=> ->; rewrite -/rank IHn.
Qed.

End FieldRank.



Section Triangular_seqmatrix.

Variable R : fieldType.


Lemma equiv_seqmatrix : forall n (M:'M[R]_(n+1)) (i j : 'I_(n+1)),
   (nth 0%R (nth nil (seqmx_of_mx M) i) j) = M i j.
Proof.
move => n M i j.
have H1 : (n.+1 = (n + 1))%N; first by rewrite addn1.
rewrite !(nth_map (cast_ord H1 (@ord0 n))).
 - by rewrite -!enumT ?nth_ord_enum.
 - by rewrite -enumT size_enum_ord (ltn_ord j).
 - by rewrite -enumT size_enum_ord (ltn_ord i).
 - by rewrite size_enum_ord (ltn_ord j).
 - by rewrite size_enum_ord (ltn_ord i).
Qed.

Definition lower_triangular_seq n (M:seqmatrix R) :=
  (forall k, 0 <= k < (n+1) -> (nth 0%R (nth nil M k) k) = 1%R) /\
  (forall r s, (0 <= r < (n+1)) && (0 <= s < (n+1)) && (r < s) -> (nth 0%R (nth nil M r) s) = 0%R).

Lemma eq_nat_of_ord_eq_ord  n (i j:'I_n) : (nat_of_ord i)==(nat_of_ord j) -> i==j.
Proof.
by move=>H; apply/eqP; apply : val_inj; move : (eqP H) => H1.
Qed.

Lemma ltn_nat_of_ord_eq_ord  n (i j:'I_n) : (nat_of_ord i)<(nat_of_ord j) -> (i==j)=false.
Proof.
by move=>H; apply/eqP => H1; rewrite H1 in H; move : H; rewrite ltnn.
Qed.


Lemma lower_triangular_seqE : forall n (M:'M[R]_(n+1)),
   lower_triangular_seq n (seqmx_of_mx M) <-> (lower_triangular M).
Proof.
split.
  + move => [H1 H2].
    rewrite /lower_triangular => i j ij.
    rewrite -equiv_seqmatrix.
    move : ij; rewrite leq_eqVlt; move/orP => [H3|H4].
      - rewrite !(eqP H3) H1; last first.
          - by rewrite (ltn_ord j).
        by rewrite (eq_nat_of_ord_eq_ord H3).
      - have H5 : (0 <= i < n+1) && (0 <= j < n+1) && (i < j); first by rewrite H4 (ltn_ord j) (ltn_ord i).
        rewrite (H2 i j H5).
        by rewrite (ltn_nat_of_ord_eq_ord H4).
  + move => H; split.
      - move => k0; rewrite /= => H1.
        rewrite (equiv_seqmatrix M (Ordinal H1) (Ordinal H1)) H //.
        by have -> : (Ordinal H1 == Ordinal H1); first by apply/eqP.
      - move => r s; move/andP => [H1 H2]; move : H1; rewrite /=; move/andP => [H3 H4].
         rewrite (equiv_seqmatrix M (Ordinal H3) (Ordinal H4)) H; last by rewrite /= leq_eqVlt H2 orb_true_r. 
         have -> : (Ordinal H3 == Ordinal H4) = false; last by done. 
         have H5 : (nat_of_ord (Ordinal H3))!=(nat_of_ord (Ordinal H4)); last first. 
          by rewrite -negb_true_iff. 
         by rewrite //= neq_ltn; apply/orP; left. 

Qed.

End Triangular_seqmatrix.






Section t_reorderM_equivalence.

Variable R: ringType.


Definition t_take_rows_s (s:seq nat) (M: seqmatrix R):=
   map (fun i => (nth [::] M (nth 0%N s i))) (iota 0 (size M)).

Definition t_take_row_s (s:seq nat)(l:seq R):=
  map (fun i => (nth 0 l (nth 0%N s i))) (iota 0 (size l)).


Definition t_take_columns_s (s:seq nat) (M: seqmatrix R):=
  map (t_take_row_s s) M.


Definition t_reorderM (s1 s2: seq nat)(M: seqmatrix R):=
  (t_take_columns_s s2 (t_take_rows_s s1 M)).



Variable s s1: seq nat.
Variable m n:nat.
Hypothesis H : perm_eq s (iota 0 m).
Hypothesis H1 : perm_eq s1 (iota 0 n).

Lemma t_take_rows_sE (M1:'M[R]_(m,n)) :
   (t_take_rows_s s (seqmx_of_mx M1)) = (seqmx_of_mx (row_perm (perm_of_seq H s) M1)).
Proof.
apply/seqmxP; split=> [|i Hi|i j].
  - by rewrite size_map size_iota size_map size_enum_ord.
  - rewrite /rowseqmx (nth_map 0%N); last by rewrite size_iota size_map size_enum_ord.
    rewrite nth_iota; last by rewrite size_map size_enum_ord.
    rewrite add0n size_row_seqmx; first by done.
    have H2 : m = size s; first by rewrite (perm_eq_size H) size_iota.
    rewrite H2; rewrite H2 in Hi.
    by move : (mem_nth 0%N Hi); rewrite (perm_eq_mem H) mem_iota /= add0n H2.
  - rewrite /fun_of_seqmx /rowseqmx (nth_map 0%N); last by rewrite size_iota size_map size_enum_ord.
    rewrite nth_iota; last by rewrite size_map size_enum_ord.
    rewrite add0n mxE -seqmxE /=.
    have -> : (nth 0%N s i) = ((perm_of_seq H s) i); last by done.
    by rewrite /= -perm_of_seqE fun_of_permseqE.
Qed.


Lemma t_take_columns_sE (M1:'M[R]_(m,n)) :
   (t_take_columns_s s1 (seqmx_of_mx M1)) = (seqmx_of_mx (col_perm (perm_of_seq H1 s1) M1)).
Proof.
apply/seqmxP; split=> [|i Hi|i j].
  - by rewrite size_map size_map size_enum_ord.
  - rewrite /rowseqmx (nth_map [::]); last by rewrite size_map size_enum_ord.
    rewrite size_map size_iota.
    have H2: (0 < m); first by case H3: m=>[|m1]; first by move: Hi; rewrite H3.
    rewrite (nth_map (Ordinal H2)); first by rewrite size_map size_enum_ord.
    by rewrite size_enum_ord.
  - rewrite /fun_of_seqmx /rowseqmx (nth_map [::]); last by rewrite size_map size_enum_ord.
    rewrite (nth_map 0%N) /=.
      rewrite nth_iota; last first.
        have H2: (0 < m); first by case H3: m=>[|m1]; first by move: (ltn_ord i); rewrite {2}H3.
        rewrite (nth_map (Ordinal H2)); first by rewrite size_map size_enum_ord.
        by rewrite size_enum_ord.
      rewrite add0n mxE -seqmxE /=.
      have -> : (nth 0%N s1 j) = ((perm_of_seq H1 s1) j); last by done.
      by rewrite /= -perm_of_seqE fun_of_permseqE.
    rewrite size_iota size_row_seqmx; first by rewrite (ltn_ord j).
    by rewrite (ltn_ord i).
Qed.



Definition reorder_mx (s1:'S_m)(s2:'S_n)(M:'M[R]_(m,n)):=
  col_perm s2 (row_perm s1 M).


Lemma t_reorderM_E (M1:'M[R]_(m,n)) :
   t_reorderM s s1 (seqmx_of_mx M1) = (seqmx_of_mx (reorder_mx (perm_of_seq H s) (perm_of_seq H1 s1) M1)).
Proof.
by rewrite /t_reorderM /reorder_mx t_take_rows_sE t_take_columns_sE.
Qed.


Lemma t_reorderM2_E (M1:seqmatrix R) :
  M1!= [::] -> [/\ m = size M1 & forall i, i < m -> size (rowseqmx M1 i) = n] ->
  reorder_mx (perm_of_seq H s) (perm_of_seq H1 s1) (mx_of_seqmx m n M1) = mx_of_seqmx m n (t_reorderM s s1 M1).
Proof.
move => HH H2.
by rewrite -{2}(seqmx_of_mx_mx_of_seqmx H2) t_reorderM_E mx_of_seqmx_seqmx_of_mx.
Qed.


End t_reorderM_equivalence.




